/*
 * This is a machine-generated file. Please do not edit it by hand.
 *
 * This file contains low-level information about the
 * internals of a particular version and configuration
 * of SBCL. It is used by the C compiler to create a runtime
 * support environment, an executable program in the host
 * operating system's native format, which can then be used to
 * load and run 'core' files, which are basically programs
 * in SBCL's own format.
 */
INITIALIZE_TLS(FREE_INTERRUPT_CONTEXT_INDEX, 0);
INITIALIZE_TLS(ALLOW_WITH_INTERRUPTS, LISP_T);
INITIALIZE_TLS(UNBLOCK_DEFERRABLES_ON_ENABLING_INTERRUPTS_P, NIL);
INITIALIZE_TLS(INTERRUPTS_ENABLED, LISP_T);
INITIALIZE_TLS(INTERRUPT_PENDING, NIL);
INITIALIZE_TLS(IN_WITHOUT_GCING, NIL);
INITIALIZE_TLS(GC_INHIBIT, NIL);
INITIALIZE_TLS(GC_PENDING, NIL);
INITIALIZE_TLS(STOP_FOR_GC_PENDING, NIL);
INITIALIZE_TLS(UNWEAKENED_VECTORS, NIL);
INITIALIZE_TLS(PINNED_OBJECTS, NIL);
INITIALIZE_TLS(GC_PIN_CODE_PAGES, 0);
INITIALIZE_TLS(CURRENT_CATCH_BLOCK, 0);
INITIALIZE_TLS(CURRENT_UNWIND_PROTECT_BLOCK, 0);
