x86-64-arch.o x86-64-arch.d : x86-64-arch.c genesis/sbcl.h runtime.h lispobj.h align.h \
 globals.h os.h target-os.h target-arch-os.h target-arch.h validate.h \
 thread.h genesis/symbol.h lispobj.h symbol-tls.inc align.h \
 genesis/vector.h genesis/static-symbols.h genesis/thread.h \
 gencgc-alloc-region.h genesis/vector.h interrupt.h gc-typedefs.h arch.h \
 lispregs.h target-lispregs.h genesis/regnames.h interr.h breakpoint.h \
 pseudo-atomic.h search.h var-io.h genesis/compiled-debug-info.h code.h \
 genesis/closure.h gc-assert.h core.h gc.h immobile-space.h gencgc-impl.h \
 gencgc-alloc-region.h genesis/cardmarks.h genesis/cons.h \
 genesis/weak-pointer.h hopscotch.h forwarding-ptr.inc \
 x86-arch-shared.inc
