var-io.o var-io.d : var-io.c gc-assert.h interr.h os.h genesis/sbcl.h runtime.h \
 lispobj.h align.h target-os.h target-arch-os.h target-arch.h var-io.h \
 genesis/number-types.h lispobj.h genesis/vector.h genesis/cons.h \
 genesis/compiled-debug-info.h code.h genesis/closure.h thread.h \
 globals.h genesis/symbol.h symbol-tls.inc align.h genesis/vector.h \
 genesis/static-symbols.h genesis/thread.h gencgc-alloc-region.h \
 interrupt.h validate.h gc-typedefs.h
