/*
 * This is a machine-generated file. Please do not edit it by hand.
 *
 * This file contains low-level information about the
 * internals of a particular version and configuration
 * of SBCL. It is used by the C compiler to create a runtime
 * support environment, an executable program in the host
 * operating system's native format, which can then be used to
 * load and run 'core' files, which are basically programs
 * in SBCL's own format.
 */
#ifndef SBCL_GENESIS_COMPILED_DEBUG_INFO
#define SBCL_GENESIS_COMPILED_DEBUG_INFO
#include "lispobj.h"
struct compiled_debug_info {
    lispobj header; // = word_0_
    lispobj name;
    lispobj source;
    lispobj eh_locs;
    lispobj package;
    lispobj fun_map;
    lispobj contexts;
    lispobj rest;
};
#endif
