arena.o arena.d : arena.c genesis/sbcl.h runtime.h lispobj.h align.h globals.h \
 os.h target-os.h target-arch-os.h target-arch.h gc.h gc-assert.h \
 interr.h gc-typedefs.h immobile-space.h gencgc-impl.h \
 gencgc-alloc-region.h genesis/cardmarks.h lispobj.h genesis/cons.h \
 genesis/weak-pointer.h hopscotch.h code.h genesis/closure.h \
 forwarding-ptr.inc lispregs.h target-lispregs.h genesis/regnames.h \
 genesis/arena.h genesis/gc-tables.h thread.h genesis/symbol.h \
 symbol-tls.inc align.h genesis/vector.h genesis/static-symbols.h \
 genesis/thread.h gencgc-alloc-region.h genesis/vector.h interrupt.h \
 validate.h genesis/instance.h instance.inc graphvisit.h tiny-lock.h
