/*
 * This is a machine-generated file. Please do not edit it by hand.
 *
 * This file contains low-level information about the
 * internals of a particular version and configuration
 * of SBCL. It is used by the C compiler to create a runtime
 * support environment, an executable program in the host
 * operating system's native format, which can then be used to
 * load and run 'core' files, which are basically programs
 * in SBCL's own format.
 */
#ifndef SBCL_GENESIS_STATIC_SYMBOLS
#define SBCL_GENESIS_STATIC_SYMBOLS
#ifndef __ASSEMBLER__
#include "lispobj.h"
struct static_trailer_constants {
  uword_t lisp_linkage_table;
  uword_t alien_linkage_table;
  uword_t function_layout;
  uword_t msan_xor_constant;
  uword_t text_space_addr;
  uword_t text_card_count;
  uword_t text_card_marks;
  uword_t cpu_feature_bits;
};
#endif
#define NIL_CARDTABLE_DISP $0x69
#define NIL LISPOBJ(STATIC_SPACE_START + 0xFFF97)
#define LISP_T LISPOBJ(STATIC_SPACE_START + 0xFFF1F)
#define IMMOBILE_CODEBLOB_TREE LISPOBJ(STATIC_SPACE_START + 0x81F)
#define IMMOBILE_CODEBLOB_VECTOR LISPOBJ(STATIC_SPACE_START + 0x84F)
#define DYNSPACE_CODEBLOB_TREE LISPOBJ(STATIC_SPACE_START + 0x87F)
#define LINKAGE_NAME_MAP LISPOBJ(STATIC_SPACE_START + 0x8AF)
#define FINALIZER_STORE LISPOBJ(STATIC_SPACE_START + 0x8DF)
#define FINALIZER_REHASHLIST LISPOBJ(STATIC_SPACE_START + 0x90F)
#define FINALIZERS_TRIGGERED LISPOBJ(STATIC_SPACE_START + 0x93F)
#define RUN_GC_HOOKS LISPOBJ(STATIC_SPACE_START + 0x96F)
#define STARTING_THREADS LISPOBJ(STATIC_SPACE_START + 0x99F)
#define FREE_TLS_INDEX LISPOBJ(STATIC_SPACE_START + 0x9CF)
#define TLS_SYMBOL_MAP LISPOBJ(STATIC_SPACE_START + 0x9FF)
#define N_PROFILE_SITES LISPOBJ(STATIC_SPACE_START + 0xA2F)
#define CALLBACK_WRAPPER_TRAMPOLINE LISPOBJ(STATIC_SPACE_START + 0xA5F)
#define LFLIST_TAIL_ATOM LISPOBJ(STATIC_SPACE_START + 0xAB3)
#define FREE_INTERRUPT_CONTEXT_INDEX (*)
#define FREE_INTERRUPT_CONTEXT_INDEX_tlsindex 0x188
#define ALLOW_WITH_INTERRUPTS (*)
#define ALLOW_WITH_INTERRUPTS_tlsindex 0x190
#define UNBLOCK_DEFERRABLES_ON_ENABLING_INTERRUPTS_P (*)
#define UNBLOCK_DEFERRABLES_ON_ENABLING_INTERRUPTS_P_tlsindex 0x198
#define INTERRUPTS_ENABLED (*)
#define INTERRUPTS_ENABLED_tlsindex 0x1A0
#define INTERRUPT_PENDING (*)
#define INTERRUPT_PENDING_tlsindex 0x1A8
#define IN_WITHOUT_GCING (*)
#define IN_WITHOUT_GCING_tlsindex 0x1B0
#define GC_INHIBIT (*)
#define GC_INHIBIT_tlsindex 0x1B8
#define GC_PENDING (*)
#define GC_PENDING_tlsindex 0x1C0
#define STOP_FOR_GC_PENDING (*)
#define STOP_FOR_GC_PENDING_tlsindex 0x1C8
#define UNWEAKENED_VECTORS (*)
#define UNWEAKENED_VECTORS_tlsindex 0x1D0
#define PINNED_OBJECTS (*)
#define PINNED_OBJECTS_tlsindex 0x1D8
#define GC_PIN_CODE_PAGES (*)
#define GC_PIN_CODE_PAGES_tlsindex 0x1E0
#define CURRENT_CATCH_BLOCK (*)
#define CURRENT_CATCH_BLOCK_tlsindex 0x18
#define CURRENT_UNWIND_PROTECT_BLOCK (*)
#define CURRENT_UNWIND_PROTECT_BLOCK_tlsindex 0x20
#define LAYOUT_OF_FUNCTION (lispobj)(FIXEDOBJ_SPACE_START+0x183)
#define SUB_GC_fname_index 1
#define POST_GC_fname_index 2
#define INTERNAL_ERROR_fname_index 3
#define CONTROL_STACK_EXHAUSTED_ERROR_fname_index 4
#define BINDING_STACK_EXHAUSTED_ERROR_fname_index 5
#define ALIEN_STACK_EXHAUSTED_ERROR_fname_index 6
#define HEAP_EXHAUSTED_ERROR_fname_index 7
#define UNDEFINED_ALIEN_VARIABLE_ERROR_fname_index 8
#define MEMORY_FAULT_ERROR_fname_index 9
#define UNHANDLED_TRAP_ERROR_fname_index 10
#define HANDLE_BREAKPOINT_fname_index 11
#define HANDLE_SINGLE_STEP_TRAP_fname_index 12
#define ENTER_ALIEN_CALLBACK_fname_index 13
#define ENTER_FOREIGN_CALLBACK_fname_index 14
#endif
